﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.workertable;
// ワーカーテーブル

private import std.thread;
private import std.c.time;
private import pes.system.all;
private import pes.prompt;
private import std.date;

class WorkerTermianteEvent: Error
{
	this()
	{
		super("terminate event");
	}
}

class WorkerTask: Task
{
protected:
	const uint WORK_SEC = 20;
	
	EventResource	m_event;
	
	char[]		m_name;
	uint 		m_work_max;
	uint 		m_work_bytes;
	uint		m_work_pos;
	time_t		m_last_work_time;
	
public:
	abstract ubyte[] hash();
	
	char[]	name()
	{
		return m_name;
	}
	
	void name(char[] n)
	{
		m_name = n;
	}
	
	this(mailbox id, TaskResource resource)
	{
		super(id, resource);
	}
	
	uint rate()	//	work/second
	{
		time_t cur = time(null);
		uint rate;
		
		uint diff = cur - m_last_work_time;
		
		if (diff != 0) {
			rate = (m_work_bytes / diff);
		}
		
		return rate;
	}
	
	void workMax(uint max)
	{
		m_work_max = max;
	}
	
	uint workMax()
	{
		return m_work_max;
	}
	
	void work(uint cmp)
	{
		time_t cur;
		
		m_work_pos += cmp;
		
		cur = time(null);
		if (m_last_work_time + WORK_SEC > cur) {
			m_work_bytes += cmp;
		} else {
			m_work_bytes = cmp;
			m_last_work_time = cur;
		}
	}
	
	uint ratio()
	{
		if (m_work_max == 0) {
			return 100;
		}
		
		return cast(uint)((cast(double)m_work_pos / m_work_max) * 100);
	}
	
	void pollTerminateEvent()
	in {
		assert(m_event !is null);
	}
	body {
		if (m_event.pollStartEvent(id, 1)) {
			throw new WorkerTermianteEvent;
		}
	}
	
	void terminateEvent()
	{
		m_event.startEvent(id, 1);
	}
	
	void eventResource(EventResource resource)
	{
		m_event = resource;
	}
}

class WorkerTable(T: WorkerTask): Locker!(uint, T)
{
protected:
	EventResource m_event;
	
public:
	this(SemaphoreResource resource, semaphore id)
	{
		super(resource, id);
		m_event = new EventResource;
	}
	
	override void set(uint key, T worker)
	{
		signal();
		try {
			m_event.create(key);	// create event flag
			worker.eventResource = m_event;
			
			super.set(key, worker);
		} finally {
			release();
		}
	}
	
	override void remove(uint key)
	{
		T worker;
		
		signal();
		try {
			worker = super.get(key);
			if (worker.getState != Thread.TS.TERMINATED) {
				worker.terminateEvent(); // 終了イベント
			}
			worker.wait(); // 終了待ち
			m_event.remove(worker.id);
			super.remove(key);
			//delete worker;
			worker = null;
		} finally {
			release();
		}
	}
	
	override void clear()
	{
		signal();
		try {
			foreach (T worker; values) {
				if (worker.getState != Thread.TS.TERMINATED) {
					worker.terminateEvent(); // 終了イベント
				}
			}
			foreach (uint key; keys) {
				remove(key);
			}
		} finally {
			release();
		}
	}
}


