﻿module pes.hashtable;

// ごくふつうのハッシュテーブル
//
// 要素数が増えるとテーブル分割がおきるが、
// 要素数が減ってもテーブル結合は起こらない
// 現在aaAの8倍くらいの遅さ！

private import pes.btree;
private import std.stdio;

class HashTable(K, V)
{
protected:
	BinaryTree!(K, V)[] m_table;
	uint				m_length = 0;
	
	void extent()
	{
		debug(pes_hashtable) printf("HashTable: extent %d: %d\n", m_length, m_table.length + 1);
		
		uint				next 	= m_table.length + 3;
		BinaryTree!(K, V)[] table	= new BinaryTree!(K, V)[next];
		
		for (int i = 0; i < next; ++i) {
			table[i] = new BinaryTree!(K, V);
		}
		for (int i = 0; i < m_table.length; ++i) {
			foreach (K e; m_table[i].keys) {
				table[typeid(K).getHash(&e) % next].set(e, m_table[i].get(e));
			}
		}
		m_table = table; // 置き換え
	}
	
public:
	this(uint init = 32)
	{
		m_table = new BinaryTree!(K, V)[init];
		
		for (uint i = 0; i < init; ++i) {
			m_table[i] = new BinaryTree!(K, V);
		}
	}
	
	void set(K key, V value)
	{
		//debug(pes_hashtable) printf("set %d\n", m_length);
		uint index 	= typeid(K).getHash(&key) % m_table.length;
		uint len	= m_table[index].length;
		
		m_table[index].set(key, value);
		if (m_table[index].length != len) {
			++m_length;
		}
		
		// ノードが増えすぎたら分割する
		if (m_length > (1 << (m_table.length / 8))) {
			extent();
		}
	}
	
	V get(K key)
	{
		return m_table[typeid(K).getHash(&key) % m_table.length].get(key);
	}
	
	void remove(K key)
	{
		//debug(pes_hashtable) printf("remove %d\n", m_length);
		uint index 	= typeid(K).getHash(&key) % m_table.length;
		uint len	= m_table[index].length;
		
		m_table[index].remove(key);
		if (m_table[index].length != len) {
			--m_length;
		}
	}
	
	bool exist(K key)
	{
		writef("exist hashtable\n");
		writef("%d %d \n", m_table.length, typeid(K).getHash(&key) % m_table.length);
		return m_table[typeid(K).getHash(&key) % m_table.length].exist(key);
	}
	
	K[] keys()
	{
		K[] keys;
		
		for (int i = 0; i < m_table.length; ++i) {
			keys ~= m_table[i].keys;
		}
		
		return keys;
	}
	
	V[] values()
	{
		V[] values;
		
		for (int i = 0; i < m_table.length; ++i) {
			values ~= m_table[i].values;
		}
		
		return values;
	}
	
	uint length()
	{
		return m_length;
	}
}

debug(pes_hashtable)
{
	import std.stdio;
	import std.string;
	import std.random;
	
	import pes.timer;
	int main()
	{
		auto Timer timer = new Timer;
		HashTable!(uint, char[]) hash = new HashTable!(uint, char[])(128);
		HashTable!(uint, char[]) hash2 = new HashTable!(uint, char[])(128);
		
		for (uint i = 0; i < 5000; ++i) {
			uint 	num = rand();
			char[]	str = toString(num);
			
			hash.set(num, str);
		}
		
		writef("- %d\n", hash.length);
		foreach (uint e; hash.keys) {
			if (e % 2) {
				hash.remove(e);
			}
		}
		writef("- %d\n", hash.length);
		
		int x;
		foreach (uint e; hash.keys) {
			hash2.set(e, hash.get(e));
			//writef("%d: %s - %s\n", x++, e, btree.get(e));
		}
		
		return 0;
	}
}

debug(aaA)
{
	import std.stdio;
	import std.string;
	import std.random;
	
	import pes.timer;
	int main()
	{
		auto Timer timer = new Timer;
		char[][uint]hash;
		char[][uint]hash2;
		
		for (uint i = 0; i < 1000; ++i) {
			uint 	num = rand();
			char[]	str = toString(num);
			
			hash[num] = str;
		}
		writef("- %d\n", hash.length);
		foreach (uint e; hash.keys) {
			if (e % 2) {
				hash.remove(e);
			}
		}
		writef("- %d\n", hash.length);
		//hash.rehash;
		int x;
		foreach (uint e; hash.keys) {
			hash2[e] = hash[e];
			//writef("%d: %s - %s\n", x++, e, btree.get(e));
		}
		
		return 0;
	}
}
