﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.system.semaphore;
// セマフォ
// というかmutex・・

private import std.thread;
private import pes.system.threadpool;
private import pes.system.delay;
private import pes.system.task;


public alias	uint	semaphore;

public SemaphoreResource	GlobalSemaphore;

static this()
{
	GlobalSemaphore = new SemaphoreResource;
}

private import std.stdio;

private class SemaphoreCounter
{
protected:
	long m_counter;
	uint m_thread_id;
	
public:
	this()
	{
		m_counter = 0;
	}
	
	void signal()
	{
		m_thread_id = Task.getThis.id;
		++m_counter;
	}
	
	void release()
	{
		--m_counter;
		if (m_counter == 0) {
			m_thread_id = m_thread_id.init;
		}
	}
	
	bool wait()
	{
		if (m_counter == 0) {
			return false;
		}
		
		return true;
	}
	
	bool user()
	{
		return m_thread_id == Task.getThis.id;
	}
}

class SemaphoreResource
{
protected:
	SemaphoreCounter[semaphore] 	m_semaphore;
	ThreadPool[semaphore]			m_thread_queue;
	
public:
	
	this()
	{
	}
	
	void create(semaphore id)
	{
		synchronized {
			if ((id in m_semaphore) == null) {
				m_semaphore[id] 	= new SemaphoreCounter;
				m_thread_queue[id]	= new ThreadPool;
			}
		}
	}
	
	void signal(semaphore id)
	{
		Task 	self 	= Task.getThis();
		bool	resume 	= false;
		
		do {
			synchronized(m_semaphore[id]) {
				if (m_semaphore[id].user) {
					m_semaphore[id].signal();
					break; // ok
				}
				if (m_thread_queue[id].empty || resume) {
					if (! m_semaphore[id].wait) {
						m_semaphore[id].signal();
						break;
					}
				}
				m_thread_queue[id].push(self);
			}
			self.pause();
			resume = true;
		} while (true);
	}
	
	void release(semaphore id)
	{
		Task	task;
		bool	next = false;
		
		synchronized(m_semaphore[id]) {
			m_semaphore[id].release();
			
			if ((!m_thread_queue[id].empty) && (! m_semaphore[id].wait)) {
				task 	= cast(Task)m_thread_queue[id].shift;
				if (task is null) {
					throw new TaskException("cast error Task");
				}
				next = true;
			}
		}
		if (next) {
			while (! task.tryResume) {
				delayTask(1);
			}
			Task.getThis.yield();
		}
	}
	
	semaphore[] ids()
	{
		return m_semaphore.keys;
	}
}

debug(pes_system_semaphore)
{
	import	std.stdio;
	import	std.c.time;
	
	enum: semaphore {
		SEM1,
	}
	
	class Data {
		int data;
		this (int i)
		{
			data = i;
		}
		override char[] toString()
		{
			return std.string.toString(data);
		}
	}
	
	class Task1: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100; ++i) {
				signalSemaphore(SEM1);
				writef("task1: get, release %d\n", i);
				msleep(100);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task2: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100; ++i) {
				msleep(10);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				writef("task2: get, release %d\n", i);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task3: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100; ++i) {
				msleep(2);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				writef("task3: get, release %d\n", i);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task4: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 10000; ++i) {
				signalSemaphore(SEM1);
				writef("task4: get, release %d!!\n", i);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	int main()
	{
		TaskResource	resource = new TaskResource;
		resource.semaphore.create(SEM1);
		
		Task t1 = new Task1(0, resource);
		Task t2 = new Task2(0, resource);
		Task t3 = new Task3(0, resource);
		Task t4 = new Task4(0, resource);
		
		t1.start();
		t2.start();
		t3.start();
		
		
		t1.wait();
		t2.wait();
		t3.wait();
		
		t4.start();
		t4.wait();
		
		
		return 0;
	}
}

