﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.linktask;

// リンクタスク

// ノードタスクの補助的な役目で、リンク構築に勤める
// コネクター

private import	std.thread,
				std.socket,
				std.stream;

private import	pes.system.all,
				pes.net.winny.all;

private import common.environment;
private import common.task;

private import pes.prompt;

class LinkTask: TaskBase
{
protected:
	NodeTable				m_node_table;
	ConnectionTable			m_upstream_connection;
	ConnectionTable			m_downstream_connection;
	
	override TaskId taskId()
	{
		return TaskId.LINK;
	}
	
	override void initialize()
	{
		gcLimit = gcLimit * 4;
	}
	
	override void terminate()
	{
		// nothing
	}
	
	override ReturnCode control()
	{
		ReturnCode result;
		
		while (true) {
			result = control_x();
			if (result != ReturnCode.SUCCESS) {
				break;
			}
			delayTask(10);
		}
		
		return result;
	}
	
	ReturnCode control_x()
	{
		Mail		mail;
		ReturnCode	result;
		
		/* リンクタスクメインループ */
		
		mail = waitForRequest();
		
		switch (mail.messageId) {
		case MessageId.TERMINATE:
			// 終了通知
			return ReturnCode.TERMINATE;
			
		case MessageId.LINK_CONNECT:
			// イベント開始
			startEvent(EventId.RUNNING, FunctionId.LINK_RUN);
			
			do {
				result = makeFormation();
			} while (result == ReturnCode.CONTINUE);
			
			// イベント終了
			endEvent(EventId.RUNNING, FunctionId.LINK_RUN);
			
			if (result != ReturnCode.SUCCESS) {
				// どうしてもノードリストが足りない場合
				// どうしよ
				// UIタスクに断念を送る？
				// いや、グぐる先生を使ってノードリストを収集するか？
				// とりあえずメッセージを出しとく！！
				debug(node_task) mbswritef(derr, "Link: ノード情報なし!!!\n");
			}
			
			break;
			
		default:
			break;
		}
		
		fullCollect();
		
		return ReturnCode.SUCCESS;
	}
	
	ReturnCode makeFormation()
	{
		bool 				done 	= false;
		Connection 			connection;
		NyProtocolHeader	header;
		uint				exist_count = 0;
		
		if (m_node_table.length == 0) {
			// ノードがない
			return ReturnCode.FAILURE;
		}
		
		// 自分用のノードテーブルをまわす
		foreach (NodeDetails node; m_node_table.values.sort) {
			if (m_upstream_connection.exists(node.aaKey)
				|| m_downstream_connection.exists(node.aaKey))
			{
				++exist_count;
				continue; // 既に検索リンクなノードは無視
			}
			
			if (node.isSelf(config)) {
				m_node_table.remove(node.aaKey);
				continue;
			}
			if (node.priority == 0) {
				m_node_table.remove(node.aaKey);
				continue;
			}
			
			connection = new Connection;
			
			debug(node_task) mbswritef(derr, "Link: 接続試行(%s)\n", node.inetAddress);
			
			try {
				header = connection.connect(config, node);
				
				// バージョンが違うのは怖いので今のところ切断する
//				if (header.major != NyProtocolHeader.WINNY_MAJOR
//					|| header.minor != NyProtocolHeader.WINNY_MINOR)
				if (header.minor != NyProtocolHeader.WINNY_MINOR)
				{
					connection.close(new NyLowVersion);
					node.priorityDown(NodeDetails.FluctuationLevel.HI);
					node.priorityDown(NodeDetails.FluctuationLevel.HI);
					m_node_table.remove(node.aaKey);
					continue;
				}
			} catch (SocketException e) {
				debug(node_task) mbswritef(derr, "Link: 接続エラー(%s)\n", e.toString);
				
				// ノード削除
				node.priorityDown(NodeDetails.FluctuationLevel.HI);
				node.priorityDown(NodeDetails.FluctuationLevel.HI);
				m_node_table.remove(node.aaKey);
				
				// 中断イベントチェック
				if (pollStartEvent(EventId.RUNNING, FunctionId.LINK_INTERRUPT)) {
					// 中断イベント検知
					// 中断イベントを終了させる
					endEvent(EventId.RUNNING, FunctionId.LINK_INTERRUPT);
					
					done = true;
					
					break;
				}
				
				continue;	// next
			}
			
			// 接続成功
			m_upstream_connection.set(connection.aaKey, connection);
			node.priorityDown(NodeDetails.FluctuationLevel.MIDDLE);
			
			debug(node_task) mbswritef(derr, "Link: 検索リンク(%s)\n  %s\n", node.inetAddress, header.toString);
			
			// 遅すぎるか、速すぎる場合、接続優先順位をだいぶ下げる
			if (! node.isUpstreamNode(config.speed)) {
				node.priorityDown(NodeDetails.FluctuationLevel.HI);
				node.priorityDown(NodeDetails.FluctuationLevel.HI);
			}
			
			// 最大数チェック
			if (m_upstream_connection.length >= config.upstreamConnectionMax + 1) {
				done = true;
				break;
			}
			
			// この間にノードタスク側で切断される可能性もある
			if (pollStartEvent(EventId.RUNNING, FunctionId.LINK_INTERRUPT, 100)) {
				// 中断イベント検知
				
				// 中断イベントを終了させる
				endEvent(EventId.RUNNING, FunctionId.LINK_INTERRUPT);
				
				done = true;
				
				break;
			}
		}
		
		// 起動後ウェイト＋中断監視
		if (pollStartEvent(EventId.RUNNING, FunctionId.LINK_INTERRUPT, 4000)) {
			// 中断イベント検知
			
			// 中断イベントを終了させる
			endEvent(EventId.RUNNING, FunctionId.LINK_INTERRUPT);
			return ReturnCode.SUCCESS; // 中断
		}
		
		// ノード切れチェック
		if (! done) {
			if (m_node_table.length == 0
				|| m_node_table.length <= exist_count) 
			{
				// ノード情報がなくなっちゃった
				return ReturnCode.FAILURE;
			}
			
			return ReturnCode.CONTINUE; // もういちどやってみる
		}
		
		return ReturnCode.SUCCESS;
	}
	
	
public:
	this(Environment environ, TaskResource resource)
	{
		super(MailId.LINK_REQUEST,
				MailId.LINK_RESPONSE, environ, resource);
		
		// ローカルに取っとく
		m_node_table 			= environ.nodeTable;
		m_upstream_connection	= environ.upstreamConnection;
		m_downstream_connection	= environ.downstreamConnection;
	}
}

