﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.querytask;

// クエリ操作タスク

private import 	std.thread,
				std.c.time,
				std.random,
				std.socket;
				

private import 	pes.system.all,
				pes.arraybox,
				pes.random,
				pes.conv,
				pes.variant,
				pes.net.winny.all;
				
private import 	common.task,
				common.packet,
				common.environment,
				task.uicommand,
				task.timertask;

debug(query_task)
{
	import pes.prompt;
}

class QueryTask: TaskBase
{
protected:
	const uint ASCENDING_SEARCH_TICK	= 0;
	const uint DESCENDING_SEARCH_TICK	= 1;
	const uint DIFFUSE_TICK				= 2;
	
	static const char[][] MAGIC_LIST = ["zip", "avi", "mp3", "mpg", "jpg", "txt"];
	
	uint				m_dummy_id;
	uint[char[]]		m_search_word;
	ConnectionTable		m_upstream_connection;
	ConnectionTable		m_downstream_connection;
	time_t				m_last_diffuse, m_last_cond_diffuse, m_last_query;
	uint				m_diffuse_count, m_cond_diffuse_count;
	
	override TaskId taskId()
	{
		return TaskId.QUERY;
	}
	
	override void initialize()
	{
		// スレッド優先順位をINCREASEする
//		setPriority(PRIORITY.INCREASE);
		gcLimit = gcLimit / 2;
	}
	
	override void terminate()
	{
		// nothing
	}
	
	override ReturnCode control()
	{
		ReturnCode 	result;
		
		while (true) {
			result = control_x();
			if (result != ReturnCode.SUCCESS) {
				return result;
			}
			delayTask(10);
		}
		
		return result;
	}
	
	ReturnCode control_x()
	{
		Mail 	mail;
		int		count;
		
		/* クエリタスクメインループ */
		mail = waitForRequest();
		
		switch (mail.messageId) {
		case MessageId.TERMINATE:
			return ReturnCode.TERMINATE;	/* 終了		*/
			
		case MessageId.CLOCK_TICK:
			switch ((cast(Variant)mail.data).toLong) {
			case ASCENDING_SEARCH_TICK:
				// 3秒に一度検索	
				if (m_search_word.length != 0) {
					ascendingQuery();
					if (m_downstream_connection.length != 0) {
						descendingQuery();
					}
				}
				break;
			case DESCENDING_SEARCH_TICK:
				// 条件付拡散クエリ(下流)
				if (m_search_word.length != 0) {
					if (m_downstream_connection.length != 0) {
						requestConditionalDiffusion(CommandSendOption.Target.DownstreamConnection);
					} else if (m_upstream_connection.length != 0) {
						// 下流がいなくて上流がいる場合は、上流にある程度出す (1/3)
						if (rand() % 3 == 0) {
							requestConditionalDiffusion(CommandSendOption.Target.UpstreamConnection);
						}
					}
				}
				break;
			case DIFFUSE_TICK:
				// 拡散クエリ
				diffusionQuery();
				break;
			default:
				assert(0);
			}
			break;
			
		case MessageId.ADD_SEARCH_WORD:
			ArrayBox!(char) word = cast(ArrayBox!(char))mail.data;
			if (word !is null) {
				addSearchWord(word.array);
			}
			word = null;
			break;
		case MessageId.REMOVE_SEARCH_WORD:
			ArrayBox!(char) word = cast(ArrayBox!(char))mail.data;
			if (word !is null) {
				removeSearchWord(word.array);
			}
			word = null;
			break;
		case MessageId.RECEIVE_QUERY:
			NyQuery query = cast(NyQuery)mail.data;
			if (query !is null) {
				answer(query);
			}
			query = null;
			break;
		case MessageId.RECEIVE_DIFFUSION_QUERY:
			ReceivedCommand request = cast(ReceivedCommand)mail.data;
			if (request.connection.isDiffusionResponseTime) {
				diffuse(request);
				request.connection.updateLastDiffusionResponseTime();
			}
			request.command.clearBuffer();
			request.command.destroy();
			delete request.command;
			request.command = null;
			request = null;
			break;
		case MessageId.RECEIVE_COND_DIFFUSION_QUERY:
			ReceivedCommand request = cast(ReceivedCommand)mail.data;
			/* 20sec間に5回まで返す それ以外は無視 */
			if (m_cond_diffuse_count == 0) {
				conditionalDiffuse(request);
				m_last_cond_diffuse = time(null);
				++m_diffuse_count;
			} else if (m_cond_diffuse_count <= 5) {
				conditionalDiffuse(request);
				++m_cond_diffuse_count;
			} else {
				time_t cur = time(null);
				if (cur - m_last_cond_diffuse > 20) {
					conditionalDiffuse(request);
					m_last_cond_diffuse		= cur;
					m_cond_diffuse_count	= 1;
				} else {
					// ここで終わり
				}
			}
			request.command.clearBuffer();
			request.command.destroy();
			delete request.command;
			request.command = null;
			request = null;
			break;
			
		default:
			break;
		}
		
		fullCollect();
		
		return ReturnCode.SUCCESS;
	}
	
	// 拡散クエリチェック+送信
	void diffusionQuery()
	{
		uint downs	= environ.downstreamConnection.length;
		uint ups	= environ.upstreamConnection.length;
		uint datas	= environ.ulDataConnection.length + environ.dlDataConnection.length;
		bool restrict;
		
		if (datas > config.downloaderMax) {
			restrict = true; // 過度のデータ転送中は拡散頻度を下げる 1/2固定
		} else {
			restrict = false;
		}
		
		if (ups) {
			foreach (Connection e; environ.upstreamConnection.values) {
				if (e.isDiffusionRequestTime) {
					if (!restrict || rand() % 2 == 0) {
						requestDiffusion(CommandSendOption.Target.UpstreamConnection, e);
					}
					e.updateLastDiffusionRequestTime();
					return; // 負荷分散化のために1コネクションずつ時間を置く
				}
			}
		}
		if (downs) {
			foreach (Connection e; environ.upstreamConnection.values) {
				if (e.isDiffusionRequestTime) {
					if (!restrict || rand() % 2 == 0) {
						requestDiffusion(CommandSendOption.Target.DownstreamConnection, e);
					}
					e.updateLastDiffusionRequestTime();
					return; // 負荷分散化のために1コネクションずつ時間を置く
				}
			}
		}
	}
	
	// 拡散要求
	void requestDiffusion(CommandSendOption.Target target, Connection connection)
	{
		CommandSendOption option = new CommandSendOption(target, connection.nodeDetails.inetAddress);
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
				Mail.makeRequest(MessageId.SEND_COMMAND,
				makeSendCommandRequest(option, new NyDiffusionRequest)));
		
		debug(query_task_d) mbswritef(derr, "Query: 拡散クエリ転送要求送信\n");
	}
	
	// 下方向クエリ
	void descendingQuery()
	in {
		assert(m_search_word.length != 0);
	}
	body {
		if (m_downstream_connection.length == 0) {
			return;	// 下流検索リンクなし
		}
		char[]	keyword;
		
		if (environ.nodeTable.length < config.nodeMax * 0.025) {
			// ほんとうにノードが少ない場合は、
			// ヒットクエリからノード情報を取るために
			// 特殊な検索ワードを使う
			if (rand() % 2) {
				keyword = MAGIC_LIST[rand() % MAGIC_LIST.length];
			} else {
				keyword = m_search_word.keys[rand() % m_search_word.length];
			}
		} else {
			// 通常は登録されている検索ワード
			if (rand() % 3 != 0) {
				keyword = m_search_word.keys[rand() % m_search_word.length];
			} else {
				// 検索ワードのみ
				char[][] words;
				
				foreach (char[] word; m_search_word.keys) {
					if (word.length != 0 && word[0] != '%') {
						words ~= word;
					}
				}
				if (words.length == 0) {
					// キーワードなし
					keyword = m_search_word.keys[rand() % m_search_word.length];
				} else {
					keyword = words[rand() % words.length];
				}
			}
		}
		
		CommandSendOption option = new CommandSendOption(CommandSendOption.Target.DownstreamConnection, null);
		sendRequestTo(MailId.CONNECTION_REQUEST, 
			Mail.makeRequest(MessageId.SEND_COMMAND, 
				makeSendCommandRequest(option, makeQuery(keyword, true))));
	}
	
	// 上方向クエリ
	void ascendingQuery(bool donstream_flag = false)
	in {
		assert(m_search_word.length != 0);
	}
	body {
		if (m_upstream_connection.length == 0) {
			return;	// 上流検索リンクなし
		}
		
		char[]	keyword;
		
		if (environ.nodeTable.length < config.nodeMax * 0.025) {
			// ほんとうにノードが少ない場合は、
			// ヒットクエリからノード情報を取るために
			// 特殊な検索ワードを使う
			if (rand() % 2) {
				keyword = MAGIC_LIST[rand() % MAGIC_LIST.length];
			} else {
				keyword = m_search_word.keys[rand() % m_search_word.length];
			}
		} else {
			// 通常は登録されている検索ワード
			if (rand() % 3 != 0) {
				keyword = m_search_word.keys[rand() % m_search_word.length];
			} else {
				// 検索ワードのみ
				char[][] words;
				
				foreach (char[] word; m_search_word.keys) {
					if (word.length != 0 && word[0] != '%') {
						words ~= word;
					}
				}
				if (words.length == 0) {
					// キーワードなし
					keyword = m_search_word.keys[rand() % m_search_word.length];
				} else {
					keyword = words[rand() % words.length];
				}
			}
		}
		
		CommandSendOption option = new CommandSendOption(CommandSendOption.Target.UpstreamConnection, null);
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
			Mail.makeRequest(MessageId.SEND_COMMAND,
			makeSendCommandRequest(option, makeQuery(keyword))));
		
		debug(query_task) mbswritef(derr, "Query: クエリ送信(keyword: %s)\n", keyword);
	}
	
	// 上方向クエリ
	void ascendingQuery(char[] keyword)
	{
		if (m_upstream_connection.length == 0) {
			return;	// 上流検索リンクなし
		}
		CommandSendOption option = new CommandSendOption(CommandSendOption.Target.UpstreamConnection, null);
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
			Mail.makeRequest(MessageId.SEND_COMMAND,
			makeSendCommandRequest(option, makeQuery(keyword))));
		
		debug(query_task) mbswritef(derr, "Query: クエリ送信(keyword: %s)\n", keyword);
	}
	
	// 条件付拡散要求
	void requestConditionalDiffusion(CommandSendOption.Target target)
	{
		char[][] words;
		
		foreach (char[] word; m_search_word.keys) {
			if (word.length != 0 && word[0] != '%') {
				words ~= word;
			}
		}
		if (words.length == 0) {
			return;	// キーワードなし
		}
		
		requestConditionalDiffusion(words[rand() % words.length], target);
	}
	
	void requestConditionalDiffusion(char[] keyword, CommandSendOption.Target target)
	{
		CommandSendOption 				option;
		
		NyConditionalDiffusionRequest 	request = new NyConditionalDiffusionRequest;
		request.keyword = keyword;
		request.queryId	= ++m_dummy_id;
		
//		option = new CommandSendOption(CommandSendOption.Target.DownstreamConnectionAll, null);
		option = new CommandSendOption(target, null);
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
				Mail.makeRequest(MessageId.SEND_COMMAND,
				makeSendCommandRequest(option, request)));
		
		debug(query_task_d) mbswritef(derr, "Query: 条件付拡散クエリ転送要求送信\n");
	}
	
	// クエリ作成
	NyQuery makeQuery(char[] keyword, bool downstream_flag = false)
	{
		NyQuery		query 		= new NyQuery;
		ViaNode[]	node;
		
		node ~= new ViaNode(config.address, config.fileSharingPort);
		
		query.queryId	= ++m_dummy_id;	// 未使用ダミー
		query.keyword 	= keyword.dup;
		query.viaNode	= node;
		query.isDownstreamQuery = downstream_flag;
		
		return query;
	}
	
	// 検索ワード登録
	void addSearchWord(char[] keyword)
	{
		time_t cur = time(null);
		
		if (keyword.length > 1) { // 最低2文字
			debug(query) mbswritef(derr, "Query: 検索ワード追加: %s\n", word);
			m_search_word[keyword.dup] = 1;
			if (m_last_query + 3 < cur) {
				// 前から3秒以上
				ascendingQuery(keyword);
				if (environ.downstreamConnection.length != 0) {
					requestConditionalDiffusion(keyword, CommandSendOption.Target.DownstreamConnection);
				}
				m_last_query = cur;
			}
		}
	}
	
	// 検索ワード削除
	void removeSearchWord(char[] keyword)
	{
		debug(query) mbswritef(derr, "Query: 検索ワード削除: %s\n", keyword);
		if ((keyword in m_search_word) != null) {
			m_search_word.remove(keyword);
		}
	}
	
	void deleteQuery(inout NyQuery query)
	{
		query.clearBuffer();
		query.destroy();
		delete query;
		query = null;
	}
	
	// クエリ回答
	void answer(NyQuery query)
	{
		int max_keys;
		
		if (query.isDiffusionQuery) {
			debug(query_task_d) 
				mbswritef(derr, 
					"Query: 拡散クエリ受信(key: %s)\n", query.keyInformation.length);
			deleteQuery(query);
			return;
		}
		
		if (query.keyword.length == 0) {
			deleteQuery(query);
			return; // 不正
		}
		
		debug(query_task) mbswritef(derr, "Query: クエリ処理\n");
		
		/* Query */
		if (query.qr == NyQuery.Qr.QUERY) {
			if (query.isDownstreamQuery) {
				debug(query_task) mbswritef(derr, "Query: 一時下降クエリ処理\n");
			}
			if (query.viaNode.length == 0) {
				// 送信元がないクエリ
				debug(query_task) mbswritef(derr, "Query: 謎クエリ(key: %s, 数:%d)\n",
					query.keyword, query.keyInformation.length);
				
				deleteQuery(query);
				return;
			}
			requestQuery(query);
		} else if (query.qr == NyQuery.Qr.RESPONSE) {
			if (query.viaNode.length == 0) {
				debug(query_task) mbswritef(derr, "Query: クエリ帰還(key: %s, 数:%d)\n",
					query.keyword, query.keyInformation.length);
				
				deleteQuery(query);
				return;
			}
			responseQuery(query);
		}
	}
	
	// 選択
	InternetAddress selectRequestNode(ViaNode[] nodes)
	{
		InternetAddress node[];
		bool found = false;
		
		foreach (Connection e; m_upstream_connection.values) {
			found = false;
			foreach (ViaNode ve; nodes) {
				if ((e.nodeDetails.inetAddress.addr == ve.address 
						&& e.nodeDetails.inetAddress.port == ve.port)
					|| (e.nodeDetails.reportAddress.addr == ve.address 
						&& e.nodeDetails.reportAddress.port == ve.port))
				{
					found = true;
					break;
				}
			}
			if (found == false) {
				node ~= e.nodeDetails.inetAddress;
			}
		}
		if (node.length == 0) {
			return null;
		}
		
		return node[rand() % node.length];
	}
	
	// ループしてるかチェック
	bool loopGraph(ViaNode[] nodes)
	{
		foreach (ViaNode e; nodes) {
			if (config.address == e.address
				&& config.fileSharingPort == e.port)
			{
				return true;
			}
		}
		
		return false;
	}
	
	// リクエスト
	void requestQuery(NyQuery query) 
	{
		debug(query_task) mbswritef(derr, "Query: リクエストクエリ(key:%s, 数: %s)\n",
			query.keyword, query.keyInformation.length);
		InternetAddress 			send_to ;
		
		send_to = selectRequestNode(query.viaNode);
		
		if (query.keyInformation.length >= 30
			|| query.viaNode.length >= 5
			|| (send_to is null && !query.isDownstreamQuery)
			|| loopGraph(query.viaNode))
		{
			// レスポンスに切り替え
			debug(query_task) mbswritef(derr, "Query: クエリ切り替え\n");
			query.qr 		= NyQuery.Qr.RESPONSE;
			responseQuery(query);
			
			return;
		}
		
		// max個のキーを要求
		Mail mail;
		uint max = 30 - query.keyInformation.length;
		UIQSearch search = new UIQSearch;
		NyKeyInformation[] array;
		
		if (query.keyword[0] == '%') {
			// hash検索
			search.key		= query.keyword[1 .. length];
			search.keyType	= UISearchKeyType.HASH;
		} else {
			search.key		= query.keyword;
			search.keyType	= UISearchKeyType.KEYWORD;
		}
		search.max = max;
		
		// キー検索
		query.keyInformation = searchFile(search);
		
		if (query.keyInformation.length >= 30 
			|| query.isDownstreamQuery)
		{
			// レスポンスに切り替え
			debug(query_task) mbswritef(derr, "Query: クエリ切り替え\n");
			query.qr 		= NyQuery.Qr.RESPONSE;
			responseQuery(query);
			
			return;
		}
		
		// レスポンスを転送する
		query.viaNode 	= query.viaNode ~ new ViaNode(config.address,
												config.fileSharingPort);
		
		CommandSendOption option = new CommandSendOption(CommandSendOption.Target.UpstreamConnection, send_to);
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
			Mail.makeRequest(MessageId.SEND_COMMAND, makeSendCommandRequest(option, query)));
	}
	
	// レスポンス
	void responseQuery(NyQuery query) 
	{
		ViaNode next;
		
		debug(query_task) mbswritef(derr, "Query: レスポンスクエリ(key:%s, 数:%d)\n",
			query.keyword, query.keyInformation.length);
		
		if (query.viaNode.length == 1) {
			next = query.viaNode[0];
			query.viaNode(null);;
		} else {
			next = query.viaNode[length-1];
			query.viaNode(query.viaNode[0 .. length-1]);
		}
		
		debug(query_task) mbswritef(derr, "Query: レスポンス転送(to: %s)\n",
			new InternetAddress(next.address, next.port));
		// レスポンスを転送する
		CommandSendOption option;
		if (query.isDownstreamQuery) {
			query.isDownstreamQuery = false;	// 一時下降クエリフラグを落とす
			
			option = new CommandSendOption(CommandSendOption.Target.UpstreamConnection, 
															new InternetAddress(next.address, next.port));
		} else {
			option = new CommandSendOption(CommandSendOption.Target.DownstreamConnection, 
															new InternetAddress(next.address, next.port));
		}
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
			Mail.makeRequest(MessageId.SEND_COMMAND,
				makeSendCommandRequest(option, query)));
	}
	
	// 条件付拡散クエリ処理
	void conditionalDiffuse(ReceivedCommand command)
	{
		int 				max;
		NyKeyInformation[]	keyinfo;
		char[]				keyword;
		UIQSearch 			search = new UIQSearch;
		uint				speed;
		NyQuery 			query 	= new NyQuery;
		
		if (command.node.speed == 0) {
			command.command.destroy();
			delete command.command;
			command.command = null;
			return;
		}
		
		NyConditionalDiffusionRequest request = new NyConditionalDiffusionRequest(command.command);
		
		debug(query_task_d) mbswritef(derr, "--- Query: 条件付拡散クエリ処理\n");
		
		keyword = request.keyword;
		
		// 遅いほうを選択
		speed = command.node.speed > config.speed ? config.speed: command.node.speed;
		
		max = cast(int)(((cast(double)speed / 40) * 128) / 2);
		if (max > 1200) {
			max = 800 + (rand() % 200);
		}
		if (max == 0) {
			max = 30;
		}
		
		search.max = max;
		if (keyword.length != 0) {
			if (keyword[0] == '%') { // 拡散クエリでhash値はないかも･･
				search.key		= keyword[1 .. length];
				search.keyType	= UISearchKeyType.HASH;
			} else {
				search.key		= request.keyword;
				search.keyType	= UISearchKeyType.KEYWORD;
			}
			keyinfo = searchFile(search);
		}
		
// Tripになんか変なデータが載っているのちょっと調べとく		
//		if (request.trip.length != 0 
//			&& keyinfo.length < max) 
//		{
//			search.key		= request.trip;
//			search.keyType	= UISearchKeyType.TRIP;
//			search.max 		= max - keyinfo.length;
//			
//			keyinfo ~= searchFile(search);
//		}
		
		ViaNode[] viaNode = new ViaNode[1];
		viaNode[0] = new ViaNode(config.address, config.fileSharingPort);
		
		query.queryId 			= request.queryId;
		query.qr				= NyQuery.Qr.QUERY;	// なぜかリクエスト
		query.isDiffusionQuery	= true;
		query.keyInformation	= keyinfo;
		query.viaNode(viaNode);
		
		debug(query_task_d) mbswritef(derr, "Query: 条件付拡散クエリ応答(数: %d, to: %s)\n",
			keyinfo.length, command.node.inetAddress.toString);
		// レスポンスを転送する
		CommandSendOption option = new CommandSendOption(CommandSendOption.Target.AllConnection,
																			command.node.inetAddress);
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
			Mail.makeRequest(MessageId.SEND_COMMAND,
				makeSendCommandRequest(option, query)));
		
		//-- null
		query = null;
		keyinfo = null;
		viaNode = null;
	}
	
	// 拡散クエリ処理
	void diffuse(ReceivedCommand command)
	{
		int 				max, cluster_max;
		char[][] 			keys;
		NyKeyInformation[]	keyinfo;
		Mail 				mail;
		UIQSearch 			search = new UIQSearch;
		uint				speed;
		NyQuery 			query  = new NyQuery;
		
		debug(query_task_d) mbswritef(derr, "--- Query: 拡散クエリ処理\n");
		
		if (command.node.speed == 0) {
			command.command.destroy();
			delete command.command;
			command.command = null;
			return;
		}
		
		// 遅いほうを選択
		speed = command.node.speed > config.speed ? config.speed: command.node.speed;
		
		// このあたりはポエニー的調節
		max = cast(int)(((cast(double)speed / 40) * 128) / 2);
		if (max > 1200) {
			max = 400 + (rand() % 500);
		}
		if (max == 0) {
			max = 30;
		}
		
		keys 		= command.node.clusteringKey();
		cluster_max = max / 2;
		
		environ.keyTable.signal();
		keyinfo = null;
		try {
			foreach (char[] e; keys) {
				if (e.length == 0) {
					continue;
				}
				if (e[0] == '%') {
					// hash検索
					search.key		= e[1 .. length];
					search.keyType	= UISearchKeyType.HASH;
				} else {
					search.key		= e;
					search.keyType	= UISearchKeyType.KEYWORD;
				}
				search.max = cluster_max - keyinfo.length;
				
				keyinfo ~= searchFile(search);
				
				if (keyinfo.length >= cluster_max) {
					break;
				}
			}
			max -= keyinfo.length;
			
			search.key 		= "."; // .を含むファイルは多い
			search.keyType	= UISearchKeyType.KEYWORD;
			search.max		= max;
			keyinfo			~= searchFile(search);
		} finally {
			environ.keyTable.release();
		}
		
		ViaNode[] viaNode = new ViaNode[1];
		viaNode[0] = new ViaNode(config.address, config.fileSharingPort);
		
		query.queryId 			= m_dummy_id++;
		query.qr				= NyQuery.Qr.QUERY;	// なぜかリクエスト
		query.isDiffusionQuery	= true;
		query.keyInformation	= keyinfo;
		query.viaNode(viaNode);
		
		debug(query_task_d) mbswritef(derr, "Query: 拡散クエリ応答(数: %d, to: %s)\n",
			keyinfo.length, command.node.inetAddress.toString);
		// レスポンスを転送する
		CommandSendOption option = new CommandSendOption(CommandSendOption.Target.AllConnection,
																			command.node.inetAddress);
		
		sendRequestTo(MailId.CONNECTION_REQUEST,
			Mail.makeRequest(MessageId.SEND_COMMAND,
				makeSendCommandRequest(option, query)));
		
		
		//-- null
		query = null;
		keyinfo = null;
		viaNode = null;
	}
	
	NyKeyInformation[] searchFile(UIQSearch search)
	{
		NyKeyInformation[]	result;
		ubyte[] hash;
		
		if (search.key.length != 0) {
			switch (search.keyType) {
			case UISearchKeyType.KEYWORD:
				result = environ.keyTable.searchNameRandom(search.key, search.max);
				break;
			case UISearchKeyType.TRIP:
				result = environ.keyTable.searchTripRandom(search.key, search.max);
				break;
			case UISearchKeyType.HASH:
				if (search.key.length % 2 == 0) {
					hash	= hexStringToBinary(search.key);
					result 	= environ.keyTable.searchHash(hash, search.max);
					hash.length = 0;
				}
				break;
			}
		}
		debug(query_task) mbswritef(derr, "Query: キー検索(key:%s, hit:%d)\n", search.key, result.length);
		
		return result;
	}
	
public:
	this(Environment environ, TaskResource resource)
	{
		super(MailId.QUERY_REQUEST,
		MailId.QUERY_RESPONSE, environ, resource);
		m_upstream_connection	= environ.upstreamConnection;
		m_downstream_connection	= environ.downstreamConnection;
		
		// 上流検索タイマーー 3sec
		setMessageTimer(this, 3, MessageTimer.TTL.ANY, new VLong(ASCENDING_SEARCH_TICK));
		
		// 下流検索タイマーー 15sec
		setMessageTimer(this, 15, MessageTimer.TTL.ANY, new VLong(DESCENDING_SEARCH_TICK));
		
		// 拡散チェックタイマー 5sec
		setMessageTimer(this, 5, MessageTimer.TTL.ANY, new VLong(DIFFUSE_TICK));
	}
}
