﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.uitask;
/*
 * UI タスク
 *
 * by C.xxxxxxxx
 * 
 */
 
private import	std.stdio,
				std.thread,
				std.string,
				std.socket,
				std.c.time,
				std.gc;

private import	pes.system.all,
 				pes.arraybox,
 				pes.conv,
 				pes.variant,
				pes.net.winny.all,
				pes.data.stt,
				pes.exception;

private import 	common.task,
				common.environment,
				task.nodetask,
				task.keytask,
				task.querytask,
				task.controltask,
				task.linktask,
				task.connectiontask,
				task.timertask;

debug(ui_task)
{
	import pes.prompt;
}

class UserInterfaceTask: TaskBase
{
protected:
	Socket			m_socket;
	uint			m_port;
	uint			m_ny_default_port;
	
	NodeTask		m_node_task;
	KeyTask			m_key_task;
	QueryTask		m_query_task;
	ControlTask		m_control_task;
	TimerTask		m_timer_task;
	LinkTask		m_link_task;
	ConnectionTask	m_connection_task;
	
	/* UIメインループ */
	override int mainLoop()
	{
		ReturnCode result;
		
		m_thread = Thread.getThis();
		
		initialize();
		
	_MAIN_LOOP:
		while (true)
		{
			try {
				result = control();
			} catch (Exception e) {
				log("control 例外: " ~ e.toString);
				result = ReturnCode.FAILURE;
			} finally {
				terminate();
			}
			
			switch (result) {
			case ReturnCode.TERMINATE:
				log("Terminate");
				result = ReturnCode.SUCCESS;
				break _MAIN_LOOP;
			
			case ReturnCode.CONTINUE:
				log("Continue");
				try {
					initialize();
				} catch (Exception e) {
					log("initialize 例外: " ~ e.toString);
					result = ReturnCode.FAILURE;
					break _MAIN_LOOP;
				}
				continue _MAIN_LOOP;
				
			case ReturnCode.FAILURE:
				break _MAIN_LOOP;
			case ReturnCode.SUCCESS:
				break _MAIN_LOOP;
			// default: exception
			}
		}
		
		return cast(int)result;
	}
	
	override TaskId taskId()
	{
		return TaskId.UI;
	}
	
	
	/*-------------------------------------------
	 *		タスク起動処理
	 *-------------------------------------------*/
	
	// 全タスク起動
	bool startTask()
	{
		debug(ui_task)mbswritef(derr, "-- タスク起動中....\n");
		
		// スレッド生成
		
		m_node_task				= new NodeTask(environ, resource);
		m_key_task				= new KeyTask(environ, resource);
		m_query_task			= new QueryTask(environ, resource);
		m_control_task			= new ControlTask(environ, resource);
		m_timer_task			= new TimerTask(environ, resource);
		m_link_task				= new LinkTask(environ, resource);
		m_connection_task		= new ConnectionTask(environ, resource);
		
		/* 起動 */
		
		// タイマー
		startEvent(EventId.TASK_START, TaskId.TIMER);
		m_timer_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.TIMER, 30000)) {
			throw new Exception("--Timerタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Timerタスク起動 %08x\n", m_timer_task.id);
		
		// キー
		startEvent(EventId.TASK_START, TaskId.KEY);
		m_key_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.KEY, 30000)) {
			throw new Exception("--Keyタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Keyタスク起動 %08x\n", m_key_task.id);
		
		// クエリ
		startEvent(EventId.TASK_START, TaskId.QUERY);
		m_query_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.QUERY, 30000)) {
			throw new Exception("--Queryタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Queryタスク起動 %08x\n", m_query_task.id);
		
		// コントロール
		startEvent(EventId.TASK_START, TaskId.CONTROL);
		m_control_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.CONTROL, 30000)) {
			throw new Exception("--Controlタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Controlタスク起動 %08x\n", m_control_task.id);
		
		// コネクション
		startEvent(EventId.TASK_START, TaskId.CONNECTION);
		m_connection_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.CONNECTION, 30000)) {
			throw new Exception("--Connectionタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Connectionタスク起動 %08x\n", m_connection_task.id);
		
		// リンク
		startEvent(EventId.TASK_START, TaskId.LINK);
		m_link_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.LINK, 30000)) {
			throw new Exception("--Linkタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Linkタスク起動 %08x\n", m_link_task.id);
		
		// ノード
		startEvent(EventId.TASK_START, TaskId.NODE);
		m_node_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.NODE, 30000)) {
			throw new Exception("--Nodeタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Nodeタスク起動 %08x\n", m_node_task.id);
		
		/* 全タスク起動通知 */
		startEvent(EventId.TASK_START, TaskId.ALL);
		debug(ui_task)mbswritef(derr, "-- 全タスク起動\n");
		
		return true;
	}
	
	// 全タスク終了
	void endTask()
	{
		debug(ui_task)mbswritef(derr, "-- 全タスク停止中....\n");
		Mail end_message = Mail.makeRequest(MessageId.TERMINATE, null);
		
		sendRequestTo(MailId.NODE_REQUEST, end_message);
		sendRequestTo(MailId.LINK_REQUEST, end_message);
		sendRequestTo(MailId.CONNECTION_REQUEST, end_message);
		sendRequestTo(MailId.TIMER_REQUEST, end_message);
		sendRequestTo(MailId.KEY_REQUEST, end_message);
		sendRequestTo(MailId.QUERY_REQUEST, end_message);
		sendRequestTo(MailId.CONTROL_REQUEST, end_message);
		
		startEvent(EventId.TASK_START, TaskId.ALL);
//		pollEndEvent(EventId.TASK_START, 	TaskId.KEY|TaskId.QUERY|TaskId.CONTROL|
//											TaskId.CACHE|TaskId.LINK|TaskId.NODE|TaskId.CONNECTION, 5000);
		m_node_task.wait();
		m_link_task.wait();
		m_connection_task.wait();
		m_timer_task.wait();
		m_key_task.wait();
		m_query_task.wait();
		m_control_task.wait();
		
		delete m_node_task;
		delete m_key_task;
		delete m_query_task;
		delete m_control_task;
		delete m_timer_task;
		delete m_link_task;
		delete m_connection_task;
		
		resource.mailBox.remove(); // 全て削除
		environ.terminate();
		
		m_resource = null;
//		delete m_resource;
		
		debug(ui_task)mbswritef(derr, "-- 全タスク停止\n");
	}
	
	// 死んだとき用
	bool checkEndTask()
	{
		TaskId id;
		
		for (int i = 0; i < TaskId.sizeof * 8; ++i) {
			id = cast(TaskId)(1 << i);
			
			if (pollEndEvent(EventId.TASK_START, id)) {
				return true;
			}
		}
		
		return false;
	}
	
	
	/* タスクリソース作成	*/
	void createResource()
	{
		Exception exception;
		Configuration config = new Configuration;
		m_resource = new TaskResource;
		
		try {
			loadConfiguration(config);
		} catch (Exception e) {
			exception = e;
		}
		
		// メールボックス
		foreach (MailId id; g_mail_ids) {
			resource.mailBox.create(id);
		}
		
		// イベントフラグ
		foreach (EventId id; g_event_ids) {
			resource.eventFlag.create(id);
		}
		
		// セマフォ
		foreach (SemaphoreId id; g_semaphore_ids) {
			resource.semaphore.create(id);
		}
		
		environ.initialize(m_resource.semaphore, config);
		if (exception) {
			throw exception;
		}
	}
	
public:
	this(uint uiport, uint ny_port)
	{
		super(MailId.UI_REQUEST, MailId.UI_RESPONSE, getEnvironment(), null);
		m_port				= uiport;
		m_ny_default_port 	= ny_port;
	}
	
	// コネクションタスク、ノードタスクを再起動
	bool rebootNode()
	{
		Mail mail;
		
		endEvent(EventId.TASK_START, TaskId.NODE);
		sendRequestTo(MailId.NODE_REQUEST, Mail.makeRequest(MessageId.REBOOT, null));
		
		mail = waitForResponse(); // 同期通信(ブロック)
		
		if (mail.status == RequestStatus.OK) {
			// OK
			endEvent(EventId.TASK_START, TaskId.CONNECTION);
			sendRequestTo(MailId.CONNECTION_REQUEST, Mail.makeRequest(MessageId.REBOOT, null));
			waitForStartEvent(EventId.TASK_START, TaskId.CONNECTION);
			waitForStartEvent(EventId.TASK_START, TaskId.NODE);
		} else {
			// NG
			startEvent(EventId.TASK_START, TaskId.NODE);
			return false;
		}
		
		return true;
	}
	
	void addSearch(char[] key)
	{
		// クエリタスクへ検索ワード登録要求
		sendRequestTo(MailId.QUERY_REQUEST, 
				Mail.makeRequest(MessageId.ADD_SEARCH_WORD, new ArrayBox!(char)(key)));
	}
	
	void removeSearch(char[] key)
	{
		// クエリタスクへ検索ワード削除要求
		sendRequestTo(MailId.QUERY_REQUEST, 
			Mail.makeRequest(MessageId.REMOVE_SEARCH_WORD, new ArrayBox!(char)(key)));
	}
	
	void addDownload(DownloadItem item)
	{
		// コントロールタスクへダウンロード登録要求
		sendRequestTo(MailId.CONTROL_REQUEST, 
				Mail.makeRequest(MessageId.ADD_DOWNLOAD, item));
	}
	
	void removeDownload(DownloadItem item)
	{
		// コントロールタスクへダウンロード削除要求
		sendRequestTo(MailId.CONTROL_REQUEST, 
				Mail.makeRequest(MessageId.REMOVE_DOWNLOAD, item));
	}
	
	void decodeCache(DownloadItem item)
	{
		// コントロールタスクへキャッシュデコード要求
		sendRequestTo(MailId.CONTROL_REQUEST, 
				Mail.makeRequest(MessageId.DECODE_CACHE, item));
	}
	
	void removeCache(DownloadItem item)
	{
		// コントロールタスクへキャッシュ削除要求
		sendRequestTo(MailId.CONTROL_REQUEST, 
				Mail.makeRequest(MessageId.REMOVE_CACHE, item));
	}
	
	void loadConfiguration(Configuration config)
	{
		STTList basic = loadSTT(config.stt, "t");
		
		config.fileSharingPort 	= basic["port"].toLong;
		config.speed			= basic["speed"].toLong;
		config.clusteringKey(0, basic["group"][0].toString);
		config.clusteringKey(1, basic["group"][1].toString);
		config.clusteringKey(2, basic["group"][2].toString);
		config.downloadFolder	= basic["DownloadFolder"].toString;
		config.uploadFolder		= basic["UploadFolder"].toString;
		config.cacheFolder		= basic["CacheFolder"].toString;
		
		mbswritef(derr, "%s\n%s\n%s\n", config.downloadFolder, config.uploadFolder, config.cacheFolder);
	}
	
	void saveConfiguration()
	{
		// new
		
		STTList root	= new STTNode("root");
		STTList poeny	= root.setChild(new STTNode("poeny"));
		STTList basic	= poeny.setChild(new STTNode("basic config"));
		STTList group;
		
		basic.setChild(new STTLeaf("port", Variant.make(cast(long)config.fileSharingPort)));
		basic.child.bottom.setNext(new STTLeaf("speed", Variant.make(cast(long)config.speed)));
		
		group = basic.child.bottom.setNext(new STTNode("group"));
		group.setChild(new STTLeaf("_", Variant.make(config.clusteringKey(0))));
		group.child.bottom.setNext(new STTLeaf("_", Variant.make(config.clusteringKey(1))));
		group.child.bottom.setNext(new STTLeaf("_", Variant.make(config.clusteringKey(2))));
		
		basic.child.bottom.setNext(new STTLeaf("DownloadFolder", Variant.make(config.downloadFolder)));
		basic.child.bottom.setNext(new STTLeaf("UploadFolder", Variant.make(config.uploadFolder)));
		basic.child.bottom.setNext(new STTLeaf("CacheFolder", Variant.make(config.cacheFolder)));
		
		saveSTT(root, "t", config.stt);
		
/* update version
		STTList stt = loadSTT(config.stt, "t");
		
		stt["port"].assign();
		stt["speed"].assign(Variant.make(cast(long)config.speed));
		stt["group"][0].assign(Variant.make(config.clusteringKey(0)));
		stt["group"][1].assign(Variant.make(config.clusteringKey(1)));
		stt["group"][2].assign(Variant.make(config.clusteringKey(2)));
		
		saveSTT(stt, "t", config.stt);
*/
	}
	
	void loadDownloadItems()
	{
		STTList[] items = loadSTT(config.downloadList, "t").findNodes("Item");
		foreach (STTList e; items) {
			DownloadItem item = new DownloadItem(
				hexStringToBinary(e["hash"].toString),
				cast(uint)e["size"].toLong,
				e["name"].toString);
			
			try {
				CacheHeader header = environ.cacheHeader.get(item.hash, item.dataSize);
				item.mergeBitmap(header.bitmap);
			} catch (NotFound) {
				// do nothing
			}
			environ.downloadList.set(item.hash, item);
		}
	}
	
	void saveDownloadItems()
	{
		STTList root = new STTNode("root");
		STTList download_items = root.setChild(new STTNode("Download Items"));
		
		environ.downloadList.signal();
		try {
			STTNode item = null;
			STTLeaf leaf = null;
			int i;
			foreach (DownloadItem e; environ.downloadList.values) {
				if (e.isComplete) {
					continue;
				}
				item = new STTNode("Item");
				leaf = new STTLeaf("hash", binaryToHexString(e.hash));
				
				leaf.bottom.setNext(new STTLeaf("size", Variant.make(cast(long)e.dataSize)));
				leaf.bottom.setNext(new STTLeaf("name", e.name));
				
				item.setChild(leaf);
				if (i == 0) {
					download_items.setChild(item);
				} else {
					download_items.child.bottom.setNext(item);
				}
				++i;
			}
		} finally {
			environ.downloadList.release();
		}
		saveSTT(root, "t", config.downloadList);
	}
	
	void rebootTask()
	{
		endTask();
		createResource();
		startTask();
		startEvent(EventId.TASK_START, TaskId.ALL);
	}
}


